"""Sphinx configuration."""  # noqa: INP001

import functools
import inspect
from datetime import UTC, datetime
from importlib.metadata import metadata
from pathlib import Path

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_design",
    "nbsphinx",
]

project = "fastabx"
author = metadata(project)["Author"]
copyright = f"{datetime.now(tz=UTC).year}, {author}"  # noqa: A001
version = metadata(project)["Version"]
release = version

autodoc_typehints = "description"
exclude_patterns = ["build"]
html_theme = "furo"


class SourceCodeError(ValueError):
    """Some part of the source code cannot be found."""


@functools.cache
def linkcode_package() -> Path:
    """Path to the source of the package."""
    pkg = inspect.getsourcefile(__import__(project))
    if pkg is None:
        raise SourceCodeError
    return Path(pkg).parent


def linkcode_resolve(domain: str, info: dict) -> str | None:
    """Return the URL to source code."""
    if domain != "py" or not info["module"]:
        return None
    pkg = linkcode_package()
    module = __import__(info["module"], fromlist=[""])
    obj = module
    for part in info["fullname"].split("."):
        obj = getattr(obj, part)
    obj = inspect.unwrap(obj)
    if isinstance(obj, property):
        obj = obj.fget
    elif isinstance(obj, functools.cached_property):
        obj = obj.func
    fn = inspect.getsourcefile(obj)
    if fn is None:
        raise SourceCodeError
    file = str(Path(fn).relative_to(pkg))
    source, start = inspect.getsourcelines(obj)
    end = start + len(source) - 1
    version = "main"  # To update to find correct version
    return f"https://github.com/bootphon/fastabx/blob/{version}/src/fastabx/{file}#L{start}-L{end}"



"""Pytest configuration."""

import pytest


def pytest_addoption(parser: pytest.Parser) -> None:
    """CLI arguments."""
    parser.addoption("--item", action="store", default=None, help="Path to the item file")
    parser.addoption("--features", action="store", default=None, help="Path to the features directory")


def pytest_configure(config: pytest.Config) -> None:
    """Global configuration."""
    config.reference_scores = {  # HuBERT base L11
        "triphone-dev-clean.item": {
            ("within", "within"): 0.03074,
            ("across", "within"): 0.03777,
        },
        "phoneme-dev-clean.item": {
            ("within", "within"): 0.01579,
            ("across", "within"): 0.02216,
            ("within", "any"): 0.07738,
            ("across", "any"): 0.08357,
        },
    }
    config.distance = "cosine"
    config.max_size_group = 50
    config.max_x_across = 10
    config.seed = 0
    config.frequency = 50



# /// script
# requires-python = ">=3.12"
# dependencies = [
#    "tqdm>=4.67.1",
#    "torch>=2.6",
#    "numpy>=2.2",
#    "h5features==1.4.1",
# ]
# ///
# ruff: noqa: EM101, EM102, TRY003
"""Utility to convert h5features to torch tensors, and back, intended to be used alongside ABXpy."""

import argparse
from pathlib import Path

import h5features
import numpy as np
import torch
from tqdm import tqdm


def torch_to_h5features(root: Path, dest: Path, step: float, group: str = "features") -> None:
    """Convert a list of torch tensors to a single h5features file."""
    items, labels, features = [], [], []
    for path in sorted(root.glob("*.pt")):
        feats = torch.load(path, map_location="cpu").squeeze().numpy(force=True)
        if feats.ndim != 2:  # noqa: PLR2004
            raise ValueError(path)
        times = np.arange(step / 2, feats.shape[0] * step, step, dtype=np.float64)
        items.append(path.stem)
        labels.append(times)
        features.append(feats)
    dest.parent.mkdir(exist_ok=True, parents=True)
    h5features.write(dest, group, items, labels, features)


def torch_to_h5features_with_times(root_features: Path, root_times: Path, dest: Path, group: str = "features") -> None:
    """Convert lists of torch tensors (features and times) to a single h5features file."""
    items, labels, features = [], [], []
    for path in sorted(root_features.glob("*.pt")):
        feats = torch.load(path, map_location="cpu").squeeze().numpy(force=True)
        if feats.ndim != 2:  # noqa: PLR2004
            raise ValueError(path)
        times = root_times / path.name
        items.append(path.stem)
        labels.append(times)
        features.append(feats)
    dest.parent.mkdir(exist_ok=True, parents=True)
    h5features.write(dest, group, items, labels, features)


def h5features_to_torch(path: Path, root_features: Path, root_times: Path) -> None:
    """Convert a h5features file to torch tensors."""
    root_features.mkdir(exist_ok=True, parents=True)
    root_times.mkdir(exist_ok=True, parents=True)
    times, features = h5features.read(path)
    for item, feats in tqdm(features.items()):
        torch.save(torch.as_tensor(feats, dtype=torch.float32), root_features / f"{item}.pt")
        torch.save(torch.as_tensor(times[item], dtype=torch.float64), root_times / f"{item}.pt")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Conversion between h5features and torch tensors")
    subparsers = parser.add_subparsers(dest="subcommand", help="Destination format")
    parser_to_torch = subparsers.add_parser("torch", help="Convert from h5features to torch")
    parser_to_h5 = subparsers.add_parser("h5", help="Convert from torch to h5features")
    parser_to_torch.add_argument("path", type=Path, help="The input h5features file")
    parser_to_torch.add_argument("features", type=Path, help="Destination to the features directory of torch tensors")
    parser_to_torch.add_argument("times", type=Path, help="Destination to the times directory of torch tensors")
    parser_to_h5.add_argument("features", type=Path, help="The path to the features directory of torch tensors")
    parser_to_h5.add_argument("destination", type=Path, help="The output h5features file")
    parser_to_h5.add_argument("--times", type=Path, help="Path to the optional directory of times tensor")
    parser_to_h5.add_argument("--step", type=float, help="Optional time step in seconds (default: 0.02)")
    args = parser.parse_args()

    if args.subcommand == "torch":
        h5features_to_torch(args.path, args.features, args.times)
    elif args.subcommand == "h5":
        if args.step is not None and args.times is None:
            torch_to_h5features(args.features, args.destination, args.step)
        elif args.step is None and args.times is not None:
            torch_to_h5features_with_times(args.features, args.times, args.destination)
        else:
            raise ValueError("Set exactly one of --times or --step")
    else:
        raise ValueError(f"Invalid subcommand: {args.subcommand}")



"""Task module. The Task class builds all the cells for the 'by', 'on' and 'across' conditions."""

from dataclasses import dataclass
from functools import cached_property

import polars as pl
import polars.selectors as cs

from fastabx.dataset import Batch
from fastabx.verify import verify_cell

MIN_A_LEN = 2  # Minimum length of A in the ABX task.


@dataclass(frozen=True)
class Cell:
    """Individual cell of the ABX task."""

    a: Batch
    b: Batch
    x: Batch
    header: str
    description: str
    is_symmetric: bool

    def __post_init__(self) -> None:
        verify_cell((self.a.data, self.a.sizes), (self.b.data, self.b.sizes), (self.x.data, self.x.sizes))

    @cached_property
    def num_triplets(self) -> int:
        """Number of triplets in the cell."""
        nx = self.a.data.size(0) - 1 if self.is_symmetric else self.x.data.size(0)
        return self.a.data.size(0) * self.b.data.size(0) * nx

    @property
    def use_dtw(self) -> bool:
        """Whether or not to use the DTW when computing the distances for this cell."""
        return not (1 == self.a.data.size(1) == self.b.data.size(1) == self.x.data.size(1))

    def __len__(self) -> int:
        return self.num_triplets

    def __repr__(self) -> str:
        return "Cell(\n\t" + ")\n\t".join(self.description.split("), ")) + "\n)"


def cells_on_by(df: pl.LazyFrame, on: str, by: list[str]) -> pl.LazyFrame:
    """Generate the cells for the ABX task given 'on' and 'by' conditions.

    The 'by' condition is optional, the function still works if 'by' is an empty list.
    """
    df = (
        df.select([on, *by])
        .with_row_index()
        .group_by([on, *by], maintain_order=True)
        .agg(pl.col("index"))
        .with_columns(len=pl.col("index").list.len())
        .with_row_index(name="__lookup")
    )
    lookup = df.select("__lookup", "index")
    cells = df.select(cs.exclude("index"))
    cells = (
        cells.join(cells, on=by or None, suffix="_b", how="inner" if by else "cross")  # A_by == B_by
        .filter(pl.col(on) != pl.col(f"{on}_b"), pl.col("len") >= MIN_A_LEN)  # A_on != B_on,  A_len >= MIN_A_LEN
        .select(cs.exclude(f"{on}_x", "len", "len_b"))
    )
    return (
        cells.join(lookup, on="__lookup", how="left")
        .join(lookup.rename({"__lookup": "__lookup_b", "index": "index_b"}), on="__lookup_b", how="left")
        .select(cs.exclude("__lookup", "__lookup_b"))
    )


class NoAcrossError(ValueError):
    """To raise if you expect 'across' conditions but received none."""

    def __init__(self) -> None:
        super().__init__("across must be non-empty")


def cells_on_by_across(df: pl.LazyFrame, on: str, by: list[str], across: list[str]) -> pl.LazyFrame:
    """Generate the cells for the ABX task given 'on', 'by' and 'across' conditions."""
    if not across:
        raise NoAcrossError
    df = (
        df.select([on, *by, *across])
        .with_row_index()
        .group_by([on, *by, *across], maintain_order=True)
        .agg(pl.col("index"))
        .with_row_index(name="__lookup")
    )
    lookup = df.select("__lookup", "index")
    cells = df.select(cs.exclude("index"))
    cells = (
        cells.join(cells, on=by + across, suffix="_b", how="inner")  # A_{by, across} == B_{by, across}
        .filter(pl.col(on) != pl.col(f"{on}_b"))  # A_on != B_on
        .join(cells, on=by or None, suffix="_x", how="inner" if by else "cross")  # A_by, B_by == X_by
        .filter(
            pl.col(on) == pl.col(f"{on}_x"),  # A_on == X_on
            pl.col(f"{on}_b") != pl.col(f"{on}_x"),  # B_on != X_on
            *[pl.col(c) != pl.col(f"{c}_x") for c in across],  # A_across, B_across != X_across
        )
        .select(cs.exclude(f"{on}_x"))
    )
    return (
        cells.join(lookup, on="__lookup", how="left")
        .join(lookup.rename({"__lookup": "__lookup_b", "index": "index_b"}), on="__lookup_b", how="left")
        .join(lookup.rename({"__lookup": "__lookup_x", "index": "index_x"}), on="__lookup_x", how="left")
        .select(cs.exclude("__lookup", "__lookup_b", "__lookup_x"))
    )


def cell_description(on: str, by: list[str], across: list[str]) -> pl.Expr:
    """Long description of a cell."""
    cell = f"ON({on}_ax = " + pl.col(on).cast(pl.String) + f", {on}_b = " + pl.col(f"{on}_b").cast(pl.String) + ")"
    if by:
        cell += ", " + pl.concat_str([f"BY({c}_abx = " + pl.col(c).cast(pl.String) + ")" for c in by], separator=", ")
    if across:
        exprs = [f"{c}_ab = " + pl.col(c).cast(pl.String) + f", {c}_x = " + pl.col(c).cast(pl.String) for c in across]
        cell += ", " + pl.format("ACROSS({})", pl.concat_str(exprs, separator=", "))
    return cell


def cell_header(on: str, by: list[str], across: list[str]) -> pl.Expr:
    """Header of a cell."""
    return pl.concat_str([on, f"{on}_b"] + by + across + [f"{c}_x" for c in across], separator="-")



"""Benchmarking DTW."""

import argparse

import torch
from torch.utils.benchmark import Compare, Timer
from tqdm import tqdm


def benchmark_dtw(dims: list[int], num_threads: int, min_run_time: float) -> Compare:
    """Benchmark DTW on input of size (n, n) and (n, n, n, n)."""
    results = []
    for n in (pbar := tqdm(dims, leave=False)):
        pbar.set_postfix(n=n)
        x = torch.randn((n, n))
        d = torch.randn((n, n, n, n))
        sx = torch.ones(n, dtype=torch.long) * n
        t_cpu = Timer(
            stmt="dtw(x)",
            setup="from fastabx.dtw import dtw",
            globals={"x": x},
            num_threads=num_threads,
            label="2D DTW",
            sub_label=str(n),
            description="cpu",
        )
        t_gpu = Timer(
            stmt="dtw(x)",
            setup="from fastabx.dtw import dtw",
            globals={"x": x.cuda()},
            num_threads=num_threads,
            label="2D DTW",
            sub_label=str(n),
            description="gpu",
        )
        t_gpu_compiled = Timer(
            stmt="u(x)",
            setup="from fastabx.dtw import dtw; import torch; u = torch.compile(dtw, dynamic=True)",
            globals={"x": x.cuda()},
            num_threads=num_threads,
            label="2D DTW",
            sub_label=str(n),
            description="gpu - compiled",
        )
        m_cpu = t_cpu.blocked_autorange(min_run_time=min_run_time)
        m_gpu = t_gpu.blocked_autorange(min_run_time=min_run_time)
        m_gpu_compiled = t_gpu_compiled.blocked_autorange(min_run_time=min_run_time)
        results += [m_cpu, m_gpu, m_gpu_compiled]

        t_cpu_batch = Timer(
            stmt="dtw_batch(d, sx, sy, symmetric=False)",
            setup="from fastabx.dtw import dtw_batch",
            globals={"d": d, "sx": sx, "sy": sx},
            num_threads=num_threads,
            label="Batched DTW",
            sub_label=str(n),
            description="cpu",
        )
        t_gpu_batch = Timer(
            stmt="dtw_batch(d, sx, sy, symmetric=False)",
            setup="from fastabx.dtw import dtw_batch",
            globals={"d": d.cuda(), "sx": sx.cuda(), "sy": sx.cuda()},
            num_threads=num_threads,
            label="Batched DTW",
            sub_label=str(n),
            description="gpu",
        )
        t_gpu_batch_compiled = Timer(
            stmt="v(d, sx, sy, symmetric=False)",
            setup="from fastabx.dtw import dtw_batch; import torch; v = torch.compile(dtw_batch, dynamic=True)",
            globals={"d": d.cuda(), "sx": sx.cuda(), "sy": sx.cuda()},
            num_threads=num_threads,
            label="Batched DTW",
            sub_label=str(n),
            description="gpu - compiled",
        )
        m_cpu_batch = t_cpu_batch.blocked_autorange(min_run_time=min_run_time)
        m_gpu_batch = t_gpu_batch.blocked_autorange(min_run_time=min_run_time)
        m_gpu_batch_compiled = t_gpu_batch_compiled.blocked_autorange(min_run_time=min_run_time)
        results += [m_cpu_batch, m_gpu_batch, m_gpu_batch_compiled]
    return Compare(results)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Compare execution time of CPU and CUDA DTW")
    parser.add_argument("--nthreads", default=torch.get_num_threads(), type=int)
    parser.add_argument("--min-runtime", default=0.2, type=float)
    parser.add_argument("--sizes", type=int, nargs="+")
    args = parser.parse_args()

    sizes = [1, *list(range(8, 129, 8))] if args.sizes is None else args.sizes
    compare = benchmark_dtw(sizes, args.nthreads, args.min_runtime)
    compare.colorize(rowwise=True)
    compare.print()



"""Various utilities."""

import os
import sys
from dataclasses import dataclass, field
from typing import Literal

import polars as pl
import torch


def default_engine() -> Literal["cpu", "gpu"]:
    """Engine is 'gpu' if available, else 'cpu'."""
    os.environ["RMM_DEBUG_LOG_FILE"] = os.devnull  # Avoid creation of rmm_log.txt logs by RAPIDS Memory Manager
    try:
        pl.LazyFrame().collect(engine="gpu")
    except (ModuleNotFoundError, pl.exceptions.ComputeError):
        return "cpu"
    else:
        return "gpu"


@dataclass(frozen=True)
class Environment:
    """Store global environment variables."""

    device: torch.device = field(default_factory=lambda: torch.device("cuda" if torch.cuda.is_available() else "cpu"))
    engine: Literal["cpu", "gpu"] = field(default_factory=default_engine)


def load_dtw_extension() -> None:
    """Load the DTW extension.

    On Linux and Windows, we check that PyTorch has been installed with the correct CUDA version.
    """
    cuda_version = "12.4"
    if sys.platform in ["linux", "win32"] and torch.version.cuda != cuda_version:
        msg = (
            f"On Linux and Windows, the DTW extension requires PyTorch with CUDA {cuda_version}. "
            "It it not compatible with other CUDA versions, or with the CPU only version of PyTorch, "
            "even if you wanted to only use the CPU backend of the DTW. "
        )
        raise ImportError(msg)
    from . import _C  # type: ignore[attr-defined] # noqa: F401


def with_librilight_bug() -> bool:
    """Whether to reproduce the results from LibriLight ABX or not."""
    return os.getenv("FASTABX_WITH_LIBRILIGHT_BUG", "0") == "1"



"""Build the DTW PyTorch C++ extension."""

import os
import subprocess
import sys
from pathlib import Path

from setuptools import Extension, setup
from torch.utils.cpp_extension import CUDA_HOME, BuildExtension, CppExtension, CUDAExtension


def get_openmp_flags() -> tuple[list[str], list[str]]:
    """Return the compiler and linker flags for OpenMP."""
    match sys.platform:
        case "linux":
            compile_flags, link_flags = ["-fopenmp"], ["-fopenmp"]
        case "win32":
            compile_flags, link_flags = ["-openmp"], []
        case "darwin":
            brew = ["brew", "--prefix", "libomp"]
            prefix = subprocess.run(brew, check=True, text=True, capture_output=True).stdout.strip()  # noqa: S603
            compile_flags = ["-Xclang", "-fopenmp"]
            link_flags = [f"-I{prefix}/include", f"-L{prefix}/lib", "-lomp", f"-Wl,-rpath,{prefix}/lib"]
        case _:
            return [], []
    return compile_flags, link_flags


def get_extension() -> Extension:
    """Either CUDA or CPU extension."""
    use_cuda = CUDA_HOME is not None
    extension = CUDAExtension if use_cuda else CppExtension
    openmp_flags = get_openmp_flags()
    extra_compile_args = {
        "cxx": ["-fdiagnostics-color=always", "-DPy_LIMITED_API=0x030C0000", "-O3"] + openmp_flags[0],
        "nvcc": ["-O3"],
    }
    sources = [Path("src/fastabx/csrc/dtw.cpp")]
    if use_cuda:
        os.environ["TORCH_CUDA_ARCH_LIST"] = "Volta;Turing;Ampere;Ada;Hopper"
        sources.append(Path("src/fastabx/csrc/cuda/dtw.cu"))
    return extension(
        "fastabx._C",
        sources,
        extra_compile_args=extra_compile_args,
        extra_link_args=openmp_flags[1],
        py_limited_api=True,
    )


setup(
    ext_modules=[get_extension()],
    cmdclass={"build_ext": BuildExtension},
    options={"bdist_wheel": {"py_limited_api": "cp312"}},
)
